"""Generate random data to cover the 5 Unicode planes."""

import functools
import itertools
import math
import random

import tensorflow as tf

# DECOMPOSITION ################################################################

def _random_codepoint_binary(lower_plane: int=0, upper_plane: int=0x323af) -> list:
    __h = '{0:0>32b}'.format(int(random.uniform(lower_plane, upper_plane)))
    return [int(__b) for __b in __h]

def _random_codepoint_bytes(lower_plane: int=0, upper_plane: int=0x323af) -> list:
    __h = '{0:0>8x}'.format(int(random.uniform(lower_plane, upper_plane)))
    return list(bytes.fromhex(__h))

# UNICODE ######################################################################

def random_codepoint(lower_plane: int=0, upper_plane: int=0x323af, binary: bool=False) -> list:
    return (
        _random_codepoint_binary(lower_plane=lower_plane, upper_plane=upper_plane) if binary
        else _random_codepoint_bytes(lower_plane=lower_plane, upper_plane=upper_plane))

def random_character(lower_plane: int=0, upper_plane: int=0x323af) -> str:
    __h = '{0:0>8x}'.format(int(random.uniform(lower_plane, upper_plane)))
    return bytes.fromhex(__h).decode('utf-32-be', errors='replace')

# SAMPLE #######################################################################

def random_encoding(sample_size: int, lower_plane: int=0, upper_plane: int=0x323af, binary: bool=False) -> list:
    __nested = [random_codepoint(lower_plane=lower_plane, upper_plane=upper_plane, binary=binary) for _ in range(sample_size)]
    return list(itertools.chain.from_iterable(__nested))

def random_string(sample_size: int, lower_plane: int=0, upper_plane: int=0x323af) -> list:
    __nested = [random_character(lower_plane=lower_plane, upper_plane=upper_plane) for _ in range(sample_size)]
    return ''.join(__nested)

# DATASET ######################################################################

def random_dataset_of_bytes(sample_count: int, sample_size: int) -> tf.data.Dataset:
    # sample generator
    def __generator() -> iter:
        for _ in range(sample_count):
            yield [random.randint(0, 0xff) for _ in range(sample_size)]
    # wrap into a dataset
    return tf.data.Dataset.from_generator(
        generator=__generator,
        output_signature=tf.TensorSpec(shape=(sample_size,), dtype=tf.int32))

def random_dataset_of_codepoints(sample_count: int, sample_size: int, lower_plane: int=0, upper_plane: int=0x323af, binary: bool=False) -> tf.data.Dataset:
    __factor = 32 if binary else 4
    # sample generator
    def __generator() -> iter:
        for _ in range(sample_count):
            yield random_encoding(sample_size=sample_size, lower_plane=lower_plane, upper_plane=upper_plane, binary=binary)
    # wrap into a dataset
    return tf.data.Dataset.from_generator(
        generator=__generator,
        output_signature=tf.TensorSpec(shape=(__factor * sample_size,), dtype=tf.int32))

def random_dataset_of_strings(sample_count: int, sample_size: int, lower_plane: int=0, upper_plane: int=0x323af) -> tf.data.Dataset:
    # sample generator
    def __generator() -> iter:
        for _ in range(sample_count):
            yield random_string(sample_size=sample_size, lower_plane=lower_plane, upper_plane=upper_plane)
    # wrap into a dataset
    return tf.data.Dataset.from_generator(
        generator=__generator,
        output_signature=tf.TensorSpec(shape=(), dtype=tf.string))



import functools

import tensorflow as tf

import mlable.maths.ops
import mlable.shaping.axes
import mlable.shaping.hilbert
import mlable.text

# CONSTANTS ####################################################################

ANSI_REGEX = r'\x1b\[[0-9;]*[mGKHF]'

# PREPROCESS ###################################################################

def _parser_factory(features: list=[], separator: str='\x1d', targets: bool=False) -> callable:
    # select the relevant features
    __list = lambda __sample: [__sample[__f] for __f in features]
    # join them
    __join = functools.partial(tf.strings.join, separator=separator)
    # ignore if no features were given
    __list = __list if features else lambda __x: __x
    __join = __join if features else lambda __x: __x
    # inputs = targets for the autoencoders
    def __parser(inputs) -> tuple:
        return (__join(__list(inputs)), __join(__list(inputs)) if targets else None)
    # customized fn
    return __parser

def _cleaner_factory(pattern: str=ANSI_REGEX, rewrite: str='') -> callable:
    __replace = functools.partial(tf.strings.regex_replace, pattern=pattern, rewrite=rewrite, replace_global=True)
    # do nothing when the pattern is empty
    __replace = __replace if pattern else (lambda __x: __x)
    # chain the operations
    def __cleaner(inputs: tf.Tensor, targets: tf.Tensor=None) -> tuple:
        return (__replace(inputs), __replace(targets) if (targets is not None) else None)
    # customized fn
    return __cleaner

def _encoder_factory(sample_dim: int, encoding: str='UTF-8') -> callable:
    # text encoding (UTF-32-BE or UTF-8)
    __utf = functools.partial(mlable.text.encode, sample_dim=sample_dim, output_dtype=tf.uint8, output_encoding=encoding)
    # encode all
    def __encoder(inputs: tf.Tensor, targets: tf.Tensor=None) -> tuple:
        return (__utf(inputs), __utf(targets) if (targets is not None) else None)
    # customized fn
    return __encoder

def _formatter_factory(batch_dim: int, token_dim: int, order_num: int, rank_num: int) -> callable:
    # entire shape after trimming
    __shape = (batch_dim,) + rank_num * (1 << order_num,) + (token_dim,)
    # group the bytes token by token
    __group = functools.partial(mlable.shaping.axes.divide, axis=1, factor=token_dim, insert=True, right=True)
    # fold along the Hilbert curve
    __fold = functools.partial(mlable.shaping.hilbert.fold, order=order_num, rank=rank_num, axis=1)
    # enforce types
    __cast = functools.partial(tf.cast, dtype=tf.float32)
    # enforce shapes
    __reshape = functools.partial(tf.reshape, shape=__shape)
    # chain the operations
    def __formatter(inputs: tf.Tensor, targets: tf.Tensor=None) -> tuple:
        return (__cast(__reshape(__fold(__group(inputs)))), __cast(__reshape(__fold(__group(targets)))) if (targets is not None) else None)
    # customized fn
    return __formatter

def _embedder_factory(bigendian: bool=True) -> callable:
    # decompose in base 2
    __expand = functools.partial(mlable.maths.ops.expand_base, base=2, depth=8, bigendian=bigendian)
    # merge all the byte decompositions
    __merge = functools.partial(mlable.shaping.axes.merge, axis=-1, right=False)
    # embed all
    def __embedder(inputs: tf.Tensor, targets: tf.Tensor=None) -> tuple:
        return (inputs, __merge(__expand(targets)) if (targets is not None) else None)
    # customized fn
    return __embedder

# > END-TO-END #################################################################

def _wrapper(inputs: tf.Tensor, parser: callable, cleaner: callable, encoder: callable, embedder: callable, formatter: callable) -> tuple:
    # fetch the relevant features
    __inputs, __targets = parser(inputs=inputs)
    # sanitize
    __inputs, __targets = cleaner(inputs=__inputs, targets=__targets)
    # encode / tokenize
    __inputs, __targets = encoder(inputs=__inputs, targets=__targets)
    # enforce types + shapes
    __inputs, __targets = formatter(inputs=__inputs, targets=__targets)
    # represent the output in binary
    __inputs, __targets = embedder(inputs=__inputs, targets=__targets)
    # targets = inputs (in binary) for the autoencoder
    return (__inputs, __targets) if (__targets is not None) else __inputs

def factory(batch_dim: int, token_dim: int, order_num: int, rank_num: int, features: list=[], pattern: str=ANSI_REGEX, rewrite: str='', separator: str='\x1d', encoding: str='UTF-8', bigendian: bool=True, targets: bool=False) -> callable:
    __sample_dim = token_dim * (1 << (rank_num * order_num))
    # custom fn
    __parser = _parser_factory(features=features, separator=separator, targets=targets)
    __cleaner = _cleaner_factory(pattern=pattern, rewrite=rewrite)
    __encoder = _encoder_factory(sample_dim=__sample_dim, encoding=encoding)
    __formatter = _formatter_factory(batch_dim=batch_dim, token_dim=token_dim, order_num=order_num, rank_num=rank_num)
    __embedder = _embedder_factory(bigendian=bigendian)
    # actual preprocessing function
    return functools.partial(_wrapper, parser=__parser, cleaner=__cleaner, encoder=__encoder, embedder=__embedder, formatter=__formatter)



import tensorflow as tf

import mlable.metrics

# GROUP ACCURACY ##############################################################

class CategoricalGroupAccuracyFromEmbeddings(mlable.metrics.CategoricalGroupAccuracy):
    def __init__(self, decoder: callable, group: int=4, name: str='categorical_group_accuracy_from_embeddings', dtype: tf.dtypes.DType=tf.dtypes.float32, **kwargs) -> None:
        # init
        super(CategoricalGroupAccuracyFromEmbeddings, self).__init__(group=group, name=name, dtype=dtype, **kwargs)
        # decoder
        self._decoder = decoder

    def update_state(self, y_true: tf.Tensor, y_pred: tf.Tensor, sample_weight: tf.Tensor=None) -> None:
        __yt = self._decoder(y_true)
        __yp = self._decoder(y_pred)
        super(CategoricalGroupAccuracyFromEmbeddings, self).update_state(y_true=__yt, y_pred=__yp, sample_weight=sample_weight)

    def get_config(self) -> dict:
        __config = super(CategoricalGroupAccuracyFromEmbeddings, self).get_config()
        __config.update({'decoder': self._decoder})
        return __config

    @classmethod
    def from_config(cls, config: dict) -> tf.keras.metrics.Metric:
        __decoder = config.pop('decoder')
        return cls(decoder=__decoder, **config)



import functools
import math

import tensorflow as tf

import mlable.sampling
import mlable.shaping.axes
import mlable.text

# SAMPLING #####################################################################

def _sampler_factory(threshold: float=0.0, temp: float=1.0, topp: float=-1.0, topk: int=-1, bigendian: bool=True) -> callable:
    # compound bit level probabilities into byte probabilities
    __bytes = functools.partial(mlable.sampling.binary, threshold=threshold, temp=temp, topp=topp, topk=topk, bigendian=bigendian, depth=8, dtype=tf.int32)
    # sample from logits
    def __sampler(outputs: tf.Tensor) -> tf.Tensor:
        return __bytes(outputs)
    # customized fn
    return __sampler

# FORMATTING ###################################################################

def _formatter_factory() -> callable:
    # merge the token axis
    def __formatter(outputs: tf.Tensor) -> tf.Tensor:
        return mlable.shaping.axes.merge(outputs, axis=-1, right=False)
    # customized fn
    return __formatter

# DECODING #####################################################################

def _decoder_factory(drop_dim: int=0, encoding: str='UTF-32-BE', errors: str='replace') -> callable:
    __encoding_dim = 4 if '32' in encoding else 1
    # restore the leading zeros in UTF-32 encoding
    __restore = functools.partial(mlable.text.untrim, count=drop_dim, outof=__encoding_dim)
    # decode the sequence of bytes into strings
    __string = functools.partial(mlable.text.decode, encoding=encoding, errors=errors)
    # from bytes to characters
    def __decoder(outputs: tf.Tensor) -> tf.Tensor:
        return __string(__restore(outputs))
    # customized fn
    return __decoder

# > END-TO-END #################################################################

def _wrapper(outputs: tf.Tensor, sampler: callable, formatter: callable, decoder: callable, cleaner: callable) -> tuple: # masker: callable
    # sample according to the probabilities / logits
    __outputs = sampler(outputs)
    # merge the spatial dimensions
    __outputs = formatter(__outputs)
    # decode the byte values into text
    __outputs = decoder(__outputs)
    # get rid of the padding
    return cleaner(__outputs)

def factory(drop_dim: int=0, threshold: float=0.0, temp: float=1.0, topp: float=-1.0, topk: int=-1, bigendian: bool=True, encoding: str='UTF-32-BE', errors: str='replace') -> callable:
    # custom fn
    __sampler = _sampler_factory(threshold=threshold, temp=temp, topp=topp, topk=topk, bigendian=bigendian)
    __formatter = _formatter_factory()
    __decoder = _decoder_factory(drop_dim=drop_dim, encoding=encoding, errors=errors)
    # actual preprocessing function
    return functools.partial(_wrapper, sampler=__sampler, formatter=__formatter, decoder=__decoder, cleaner=mlable.text.unpad)
